import pandas as pd
import numpy as np


def right_left(field, th):
    """
    Separate geopotential field into left and right of the th line.

    Parameters
    ----------
    field (xr.DataArray): The geopotential field
    th: The direction (in degrees)

    Returns
    -------
    left, right (2 xr.DataArray): The left and right side of the geopt. field.
    """
    if th <= 180:
        return field.where((field.az <= th) | (field.az > 180 + th)), field.where(
            (field.az > th) & (field.az <= 180 + th)
        )
    else:
        return field.where((field.az <= th) & (field.az > th - 180)), field.where(
            (field.az > th) | (field.az <= th - 180)
        )


def right_left_vector(z, th):
    """
    Separate geopotential field into left and right of the th line.

    Parameters
    ----------
    z (xr.DataArray): The geopotential field
    th: The direction (in degrees)

    Returns
    -------
    left, right (2 xr.DataArray): The left and right side of the z. field.
    """

    A = pd.DataFrame([list(z.az.values)] * len(z.snapshot))  # matrix of az x snapshot
    mask = np.array(
        (A.lt(th, 0) & A.ge(th - 180, 0)) | A.ge(th + 180, 0)
    )  # Mask in 2D (az, snapshot)
    mask = np.array([mask] * len(z.r))  # Mask in 3D (r, az, snapshot)
    mask = np.swapaxes(mask, 0, 1)  # Mask in 3D (az, r, snapshot)
    R, L = z.where(mask), z.where(~mask)
    return R, L


def area_weights(field):
    """
    Computes the weights needed for the weighted mean of polar field.

    Parameters
    ----------
    field (xr.DataArray): The geopotential field

    Returns
    -------
    w (xr.DataArray): The weights corresponding to the area wrt the radius.
    """
    δ = (field.r[1] - field.r[0]) / 2
    w = (field.r + δ) ** 2 - (field.r - δ) ** 2
    return w


def B(th, geopt, SH=False, names=["snap_z900", "snap_z600"]):  # TODO: Useless?
    """
    Computes the B parameter for a point, with the corresponding snapshot of geopt at 600hPa and 900hPa

    Parameters
    ----------
    th: The direction (in degrees)
    geopt (xr.DataSet): The snapshots at both levels
    SH (bool): Set to True if the point is in the southern hemisphere
    names: names of the 900hPa and 600hPa geopt. variables in geopt.

    Returns
    -------
    B, the Hart phase space parameter for symetry.
    """
    if type(names) == str:
        z900 = geopt[names].sel(plev=900e2, method="nearest")
        print("Level " + str(z900.plev.values) + " is taken for 900hPa")
        z600 = geopt[names].sel(plev=600e2, method="nearest")
        print("Level " + str(z600.plev.values) + " is taken for 600hPa")
    else:
        z900 = geopt[names[0]]
        z600 = geopt[names[1]]

    ΔZ = z600 - z900
    ΔZ_R, ΔZ_L = right_left_vector(ΔZ, th)
    if SH:
        h = -1
    else:
        h = 1
    return (
        h
        * (
            ΔZ_R.weighted(area_weights(ΔZ_R)).mean(["r", "az"])
            - ΔZ_L.weighted(area_weights(ΔZ_L)).mean(["r", "az"])
        ).values
    )


def B_vector(th_vec, z900, z600, lat):
    """
    Computes the B parameter for a vector of points, with the corresponding snapshot of geopt at 600hPa and 900hPa

    Parameters
    ----------
    th_vec : The theta parameter for each point
    z900 : The z900 field for each point
    z600 : The z600 field for each point
    lat : The latitude of each point

    Returns
    -------
    B, the Hart phase space parameter for symetry.
    """
    ΔZ = z600 - z900
    ΔZ_R, ΔZ_L = right_left_vector(ΔZ, th_vec)
    h = np.where(lat < 0, -1, 1)
    return h * (
        ΔZ_R.weighted(area_weights(ΔZ_R)).mean(["az", "r"])
        - ΔZ_L.weighted(area_weights(ΔZ_L)).mean(["az", "r"])
    )



import numpy as np


def VT(geopt, name="snap_zg"):
    """
    Parameters
    ----------
    geopt (xr.DataArray) : The Geopotential snapshots DataArray.
        plev must be decreasing
    name (str) : Name of the geopotential snapshots variable.

    Returns
    -------
    VTL, VTU : The Hart Phase Space parameters for upper and lower thermal wind respectively.
    """
    # from sklearn.linear_model import LinearRegression
    from scipy.stats import linregress

    Z_max = geopt[name].max(["az", "r"])  # Maximum of Z at each level for each snapshot
    Z_min = geopt[name].min(["az", "r"])  # Minimum of ...
    ΔZ = Z_max - Z_min  # Fonction of snapshot & plev
    ΔZ_bottom = ΔZ.sel(plev=slice(950e2, 600e2))  # Lower troposphere
    ΔZ_top = ΔZ.sel(plev=slice(600e2, 250e2))  # Upper tropo
    X = np.log(ΔZ_bottom.plev).values.reshape(-1, 1).flatten()
    # VTL = [LinearRegression().fit(X, y).coef_[0] if not np.isnan(y).any() else np.nan for y in ΔZ_bottom.values]
    VTL = [
        linregress(X, y).slope if not np.isnan(y).any() else np.nan
        for y in ΔZ_bottom.values
    ]
    X = np.log(ΔZ_top.plev).values.reshape(-1, 1).flatten()
    # VTU = [LinearRegression().fit(X, y).coef_[0] for y in ΔZ_top.values]
    VTU = [linregress(X, y).slope for y in ΔZ_top.values]
    return VTL, VTU



import numpy as np
import pyproj


def theta(x0=120, x1=130, y0=12, y1=10):
    """
    Computes the angular direction between to points.
    0° corresponds to eastward, 90° northward, 180° westward and 270° southward.

    Parameters
    ----------
    x0: longitude coordinate of the current point
    x1: longitude coordinate of the next point
    y0: latitude coordinate of the current point
    y1: longitude coordinate of the next point

    Returns
    -------
    th (float): The directional angle between the current and the next point, in degrees.
        0 for eastward, 90 for northward.
    """
    geodesic = pyproj.Geod(ellps="WGS84")
    if (x1-x0, y1-y0) != (0,0):
        fwd_azimuth, back_azimuth, distance = geodesic.inv(x0, y0, x1, y1)
        return -1 * (fwd_azimuth - 90) % 360
    else:
        return np.nan


def theta_track(lon, lat):
    """
    Computes the angular direction for each points along a track.
    Handling the track altogether allows for treating stationnary cases as well as the end of the track.
    In stationnary cases and in the end of the track, the direction from the previous point is taken for the current point.

    Parameters
    ----------
    lon: The list of longitudes along the track
    lat: The list of latitude along the track

    Returns
    -------
    th (list): values of th along the track.
    """
    th = []
    assert len(lon) == len(lat), "The two vector do not have the same length"
    n = len(lon)

    # Compute the angle for all the points
    for i in range(
        n - 1
    ):  # Computing the direction between each point and the following
        th.append(theta(lon[i], lon[i + 1], lat[i], lat[i + 1]))
        if np.isnan(th[-1]) & (
            i != 0
        ):  # If two successive points are superimposed, we take the previous direction
            th[-1] = th[-2]
    if n > 1:
        th.append(th[-1])
        # The direction for the last point is considered the same as the point before
    else:
        th = [np.nan]
    return th


def theta_multitrack(tracks):
    """
    Compute the angular direction for every tracks in a dataset.
    All tracks must have at least two points.

    Parameters
    ----------
    tracks (pd.DataFrame): The set of TC points including columns:
        * time
        * lon
        * lat

    Returns
    -------
    thetas (list): The list of angle for each point in the dataset
    """

    assert (
        tracks.groupby("track_id").time.count().min() > 1
    ), "The dataset contains tracks with only one point."

    # Identify position of track point from the end of the track (tpos=0 for the last point)
    tracks["tpos"] = (
        tracks.sort_values("time", ascending=False)
        .groupby("track_id")
        .transform("cumcount")
    )

    lon, lat = tracks.lon.values, tracks.lat.values

    # Run theta_track as if it was all one track
    th = np.array(theta_track(lon, lat))

    # Manage last point of each track : Set same angle as the point before
    th[list((tracks.tpos == 0).values)] = th[list((tracks.tpos == 1).values)]

    return th


if __name__ == "__main__":
    import numpy as np
    import pandas as pd

    # Test theta
    assert theta(0, 1, 0, 0) == 0.0  # Eastward
    assert theta(0, 0, 0, 1) == 90.0  # Northward
    assert theta(0, -1, 0, 0) == 180.0  # Westward
    assert theta(0, 0, 0, -1) == 270.0  # Southward

    # Test theta_track
    assert theta_track([0, 1, 1, 0, 0], [0, 0, 1, 1, 0]) == [
        0.0,
        90.0,
        180,
        270.0,
        270.0,
    ]

    # Test theta_multitrack
    tracks = pd.read_csv("tests/1996.csv", index_col=False)
    assert type(theta_multitrack(tracks)) == np.ndarray
    assert theta_multitrack(tracks)[1] == 90.0

    print("All good")



from .theta import theta_multitrack
from .B import B_vector
from .VT import VT
import pandas as pd
import numpy as np


def compute_CPS_parameters(
    tracks,
    geopt,
    geopt_name="snap_zg",
    plev_name="level",
    verbose=True,
):
    """
    Computes the three (+ theta) Hart parameters for all the points in tracks.

    Parameters
    ----------
    tracks (pd.DataFrame): The set of TC points
    geopt (xr.DataSet): The geopotential snapshots associated with the tracks
        level coordinate must be in Pa.
    geopt_name (str): Provide the name of the 3D (plev, r, az) geopt snapshots variables as a string.
    plev_name (str): name of the vertical coordinate in the geopt file.

    Returns
    -------
    tracks (pd.DataFrame): The set of TC points with four new columns corresponding to the parameters
    """

    # Curate input
    ## Snapshots
    geopt = geopt.rename({plev_name: "plev"})  # Change vertical coordinate name
    geopt = geopt.where(np.abs(geopt.snap_zg) < 1e10)

    # tracks
    if "time" not in tracks.columns:
        tracks["time"] = pd.to_datetime(
            tracks.year.astype(str)
            + "-"
            + tracks.month.astype(str)
            + "-"
            + tracks.day.astype(str)
            + " "
            + tracks.hour.astype(str)
            + ":00:00"
        )

    # 1/ B computation
    if verbose:
        print("Computing B...")
    ## Select 900 & 600 hPa levels
    z900, z600 = (
        geopt[geopt_name].sel(plev=900e2, method="nearest"),
        geopt[geopt_name].sel(plev=600e2, method="nearest"),
    )
    if verbose:
        print(
            "Level "
            + str(z900.plev.values)
            + " is taken for 900hPa"
            + "\n"
            + "Level "
            + str(z600.plev.values)
            + " is taken for 600hPa"
            + "\n"
        )

    ## theta computation
    if "theta" not in tracks.columns:
        tracks = tracks.assign(theta=theta_multitrack(tracks))

    ## B computation
    tracks = tracks.assign(
        B=B_vector(tracks.theta.values, z900, z600, tracks.lat.values)
    )

    # 2/ VTL & VTU computation
    if verbose:
        print("Computing VTL & VTU...")
    geopt = geopt.sortby("plev", ascending=False)
    VTL, VTU = VT(geopt, name=geopt_name)

    # Output
    tracks = tracks.assign(VTL=VTL, VTU=VTU)

    return tracks


if __name__ == "__main__":
    import xarray as xr

    # Test theta_multitrack
    tracks = pd.read_csv("tests/Dale.csv", index_col=False)
    geopt = xr.open_dataset("tests/Dale.nc")

    df = compute_CPS_parameters(tracks, geopt)



import setuptools

setuptools.setup(
    name="CPyS",
    version="0.0.4.1",
    author="Stella Bourdin",
    author_email="stella.bourdin@lsce.ipsl.fr",
    description="A python package from compute the Hart Cyclone Phase Space parameters",
    long_description="A python package from compute the Hart Cyclone Phase Space parameters",
    # long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["numpy", "pandas", "xarray","pyproj","matplotlib"],
    #include_package_data=True,
    #package_data={"":['_data/*.csv', "_data/iho.*"]}
)


# Black formatting:
# `python -m black <directory or file(s)>`

# To generate the distribution:
# 1 / Check that wheel is up to date with `pip install --user --upgrade setuptools wheel`
# 2 / Run `python setup.py sdist bdist_wheel`

# To upload to PyPI:
# 1 / Check that twine is up-to-date `pip install --user --upgrade twine`
# 2 / Upload with `python -m twine upload --repository pypi dist/*`



import matplotlib.pyplot as plt


def plot_CPS(tracks, title=""):
    fig, axs = plt.subplots(1, 2, figsize=[10, 5])
    fig.suptitle(title)

    # Left plot (B vs. VTL)
    ## Data
    axs[0].plot(tracks.VTL, tracks.B, marker="o", color="k")
    ## x-axis
    axs[0].axvline(x=0, color="k", alpha=0.5, linestyle="--", linewidth=1)
    axs[0].set_xlabel("$-V_T^L$ / m")
    ## y-axis
    axs[0].axhline(y=10, color="k", alpha=0.5, linestyle="--", linewidth=1)
    axs[0].set_ylabel("B / m")

    # Right plot (VTU vs. VTL)
    ## Data
    axs[1].plot(tracks.VTL, tracks.VTU, marker="o", color="k")
    ## x-axis
    axs[1].axvline(x=0, color="k", alpha=0.5, linestyle="--", linewidth=1)
    axs[1].set_xlabel("$-V_T^L$ / m")
    ## y-axis
    axs[1].axhline(y=0, color="k", alpha=0.5, linestyle="--", linewidth=1)
    axs[1].set_ylabel("$-V_T^U$ / m")

    plt.tight_layout()
    plt.show()
